from pyftdi.i2c import I2cController, I2cIOError, I2cNackError


class I2cDevice:
    """Abstract base class for a device programmed using an FTDI USB-to-I²C adapter.

    To be used in cooperation with a `functionality` class such as dac.DAC.

    Should be instanciated using the `with` statement in order to release the controller
    after the communication has ended.

    """

    CONNECTION_ERR_MSSG = " ".join(
        [
            "Can't establish a connection with the slave device.",
            "Try (re)opening the bus with 'open()'.",
        ]
    )

    _controller = None
    _config_options = {}

    def __init__(self, address, url="ftdi:///1", **kwargs):
        """Init.

        Parameters
        ----------
        address : int
            I²C address of the DAC.
            You can get the value fron the method `scan_slaves` of the class
            `Controller`.
        url : str, optional
            Connection details for the adapter.
            If you don't know which value to use, call the methods `scan_masters` and
            `url_from_descriptor` from the file `i2ctools.py`.
            Otherwise the default url should work with any of the FTDI adapters.
        **kwargs
            Options to configure the I²C bus:
            * frequency : float
                I²C bus frequency, in Hertz
            * clockstretching : bool
                Enable clockstreching.
                xD5 (GPIOL1) pin should be connected to xD0 (SCK).

        """
        self._address = address
        self._url = url

        # Catch the keyword arguments for the I2cController
        if "frequency" in kwargs:
            self._config_options["frequency"] = kwargs.pop("frequency")
        if "clockstretching" in kwargs:
            self._config_options["clockstretching"] = kwargs.pop("clockstretching")

        # Forward unused keyword arguments
        super().__init__(**kwargs)

    def __enter__(self):
        """Enter a `with` block."""
        self.open()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit a `with` block."""
        self.close()

    def open(self):
        """Open and test connection to the slave device."""
        if self._controller is None:
            self._controller = I2cController()
        if not self._controller.configured:
            self._controller.configure(self._url, **self._config_options)
        self._controller.flush()
        if not self.is_open:
            self.close()
            raise ConnectionError(self.CONNECTION_ERR_MSSG)

    @property
    def is_open(self):
        """Check whether the connection to the slave device is open."""
        try:
            port = self._controller.get_port(self._address)
            port.read(0)
        except:
            return False
        else:
            return True

    def close(self):
        """Close the connection to the slave device."""
        if self._controller is not None:
            self._controller.terminate()
            self._controller = None

    def _get_port(self):
        """Obtain a port to drive the slave device and test the connection.

        Throws a `ConnectionError` exception if the connection with the device can't
        be established.

        """
        if not self.is_open:
            raise ConnectionError(self.CONNECTION_ERR_MSSG)
        return self._controller.get_port(self._address)

    def read(self, buffer_length=0):
        """Read one or more bytes from the slave device."""
        port = self._get_port()
        return port.read(buffer_length)

    def write(self, buffer):
        """Write one or more bytes to the slave device."""
        port = self._get_port()
        return port.write(buffer)
